@@ -109,6 +109,7 @@ typedef struct {
 	guint watch_id;
 	guint retry_id;
 	guint retries;
+	guint quit_blockers;
 	ConnectStep connect_step;
 	NMConnection *connection;
 	NMDBusLibreswanHelper *dbus_skeleton;
@@ -282,6 +283,25 @@ nm_libreswan_secrets_validate (NMSettingVpn *s_vpn, GError **error)
 
 /****************************************************************/
 
+static void
+block_quit (NMLibreswanPlugin *self)
+{
+	NMLibreswanPluginPrivate *priv = NM_LIBRESWAN_PLUGIN_GET_PRIVATE (self);
+	priv->quit_blockers++;
+	DEBUG ("Block quit: %d blockers", priv->quit_blockers);
+}
+
+static void
+unblock_quit (NMLibreswanPlugin *self)
+{
+	NMLibreswanPluginPrivate *priv = NM_LIBRESWAN_PLUGIN_GET_PRIVATE (self);
+	if (--priv->quit_blockers == 0)
+		g_main_loop_quit (loop);
+	DEBUG ("Unblock quit: %d blockers", priv->quit_blockers);
+}
+
+/****************************************************************/
+
 static gboolean connect_step (NMLibreswanPlugin *self, GError **error);
 static gboolean pr_cb (GIOChannel *source, GIOCondition condition, gpointer user_data);
 
@@ -465,6 +485,7 @@ check_running_cb (GPid pid, gint status, gpointer user_data)
 		ret = WEXITSTATUS (status);
 
 	DEBUG ("Spawn: child %d exited with status %d", pid, ret);
+	unblock_quit (self);
 
 	/* Reap child */
 	waitpid (pid, NULL, WNOHANG);
@@ -515,6 +536,7 @@ child_watch_cb (GPid pid, gint status, gpointer user_data)
 	priv->pid = 0;
 
 	DEBUG ("Spawn: child %d exited", pid);
+	unblock_quit (self);
 
 	if (WIFEXITED (status)) {
 		ret = WEXITSTATUS (status);
@@ -554,7 +576,8 @@ child_watch_cb (GPid pid, gint status, gpointer user_data)
 	g_clear_error (&error);
 }
 
-static gboolean do_spawn (GPid *out_pid,
+static gboolean do_spawn (NMLibreswanPlugin *self,
+                          GPid *out_pid,
                           int *out_stdin,
                           int *out_stderr,
                           GError **error,
@@ -562,7 +585,8 @@ static gboolean do_spawn (GPid *out_pid,
                           ...) G_GNUC_NULL_TERMINATED;
 
 static gboolean
-do_spawn (GPid *out_pid,
+do_spawn (NMLibreswanPlugin *self,
+          GPid *out_pid,
           int *out_stdin,
           int *out_stderr,
           GError **error,
@@ -614,6 +638,8 @@ do_spawn (GPid *out_pid,
 		*out_pid = pid;
 
 	g_ptr_array_free (argv, TRUE);
+	if (success)
+		block_quit (self);
 	return success;
 }
 
@@ -622,7 +648,7 @@ nm_libreswan_config_psk_write (NMSettingVpn *s_vpn,
                                const char *secrets_path,
                                GError **error)
 {
-	const char *pw_type, *psk, *leftid;
+	const char *pw_type, *psk, *leftid, *right;
 	int fd;
 
 	/* Check for ignored group password */
@@ -647,8 +673,13 @@ nm_libreswan_config_psk_write (NMSettingVpn *s_vpn,
 	}
 
 	leftid = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_LEFTID);
-	g_assert (leftid);
-	write_config_option (fd, "@%s: PSK \"%s\"\n", leftid, psk);
+	if (leftid) {
+		write_config_option (fd, "@%s: PSK \"%s\"\n", leftid, psk);
+	} else {
+		right = nm_setting_vpn_get_data_item (s_vpn, NM_LIBRESWAN_RIGHT);
+		g_assert (right);
+		write_config_option (fd, "%s %%any: PSK \"%s\"\n", right, psk);
+	}
 
 	close (fd);
 	return TRUE;
@@ -656,7 +687,8 @@ nm_libreswan_config_psk_write (NMSettingVpn *s_vpn,
 
 /****************************************************************/
 
-static gboolean spawn_pty (int *out_stdout,
+static gboolean spawn_pty (NMLibreswanPlugin *self,
+                           int *out_stdout,
                            int *out_stderr,
                            int *out_ptyin,
                            GPid *out_pid,
@@ -665,7 +697,8 @@ static gboolean spawn_pty (int *out_stdout,
                            ...) G_GNUC_NULL_TERMINATED;
 
 static gboolean
-spawn_pty (int *out_stdout,
+spawn_pty (NMLibreswanPlugin *self,
+           int *out_stdout,
            int *out_stderr,
            int *out_ptyin,
            GPid *out_pid,
@@ -769,6 +802,7 @@ spawn_pty (int *out_stdout,
 	if (out_pid)
 		*out_pid = child_pid;
 
+	block_quit (self);
 	return TRUE;
 }
 
@@ -1013,37 +1047,6 @@ lookup_string (GVariant *dict, const gchar *key)
 	return value;
 }
 
-static GVariant *
-route_to_gvariant (GVariant *env)
-{
-	GVariantBuilder builder;
-
-	if (!lookup_string (env, "PLUTO_PEER_CLIENT"))
-		return NULL;
-
-	g_variant_builder_init (&builder, G_VARIANT_TYPE ("au"));
-
-#define _try_add(builder, variant) \
-	G_STMT_START { \
-		GVariant *_v = (variant); \
-		\
-		if (!_v) \
-			goto fail; \
-		g_variant_builder_add_value ((builder), _v); \
-	} G_STMT_END
-	_try_add (&builder, addr4_to_gvariant (lookup_string (env, "PLUTO_PEER_CLIENT_NET")));
-	_try_add (&builder, netmask4_to_gvariant (lookup_string (env, "PLUTO_PEER_CLIENT_MASK")));
-	_try_add (&builder, addr4_to_gvariant (lookup_string (env, "PLUTO_NEXT_HOP")));
-	_try_add (&builder, g_variant_new_uint32 (0));
-	_try_add (&builder, addr4_to_gvariant (lookup_string (env, "PLUTO_MY_SOURCEIP")));
-#undef _try_add
-
-	return g_variant_builder_end (&builder);
-fail:
-	g_variant_builder_clear (&builder);
-	return NULL;
-}
-
 static void
 _take_route (GPtrArray *routes, GVariant *new, gboolean alive)
 {
@@ -1080,6 +1083,50 @@ _take_route (GPtrArray *routes, GVariant *new, gboolean alive)
 		g_variant_unref (new);
 }
 
+static void
+handle_route (GPtrArray *routes, GVariant *env, gboolean alive)
+{
+	GVariantBuilder builder;
+	const gchar *net, *mask, *next_hop, *my_sourceip;
+
+	if (!lookup_string (env, "PLUTO_PEER_CLIENT"))
+		return;
+
+	net = lookup_string (env, "PLUTO_PEER_CLIENT_NET");
+	mask = lookup_string (env, "PLUTO_PEER_CLIENT_MASK");
+	next_hop = lookup_string (env, "PLUTO_NEXT_HOP");
+	my_sourceip = lookup_string (env, "PLUTO_MY_SOURCEIP");
+
+	if (!net || !mask || !next_hop || !my_sourceip)
+		return;
+
+	if (g_strcmp0 (net, "0.0.0.0") == 0 && g_strcmp0 (mask, "0")) {
+		g_variant_builder_init (&builder, G_VARIANT_TYPE ("au"));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant ("0.0.0.0"));
+		g_variant_builder_add_value (&builder, netmask4_to_gvariant ("128.0.0.0"));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (next_hop));
+		g_variant_builder_add_value (&builder, g_variant_new_uint32 (0));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (my_sourceip));
+		_take_route (routes, g_variant_builder_end (&builder), alive);
+
+		g_variant_builder_init (&builder, G_VARIANT_TYPE ("au"));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant ("128.0.0.0"));
+		g_variant_builder_add_value (&builder, netmask4_to_gvariant ("128.0.0.0"));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (next_hop));
+		g_variant_builder_add_value (&builder, g_variant_new_uint32 (0));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (my_sourceip));
+		_take_route (routes, g_variant_builder_end (&builder), alive);
+	} else {
+		g_variant_builder_init (&builder, G_VARIANT_TYPE ("au"));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (net));
+		g_variant_builder_add_value (&builder, netmask4_to_gvariant (mask));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (next_hop));
+		g_variant_builder_add_value (&builder, g_variant_new_uint32 (0));
+		g_variant_builder_add_value (&builder, addr4_to_gvariant (my_sourceip));
+		_take_route (routes, g_variant_builder_end (&builder), alive);
+	}
+}
+
 static gboolean
 handle_callback (NMDBusLibreswanHelper *object,
                  GDBusMethodInvocation *invocation,
@@ -1173,9 +1220,9 @@ handle_callback (NMDBusLibreswanHelper *object,
 
 	/* This route */
 	if (g_strcmp0 (verb, "route-client") == 0 || g_strcmp0 (verb, "route-host"))
-		_take_route (priv->routes, route_to_gvariant (env), TRUE);
+		handle_route (priv->routes, env, TRUE);
 	else if (g_strcmp0 (verb, "unroute-client") == 0 || g_strcmp0 (verb, "unroute-host"))
-		_take_route (priv->routes, route_to_gvariant (env), FALSE);
+		handle_route (priv->routes, env, FALSE);
 
 	/* Routes */
 	g_variant_builder_init (&builder, G_VARIANT_TYPE ("aau"));
@@ -1390,7 +1437,7 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 		priv->connect_step++;
 
 	case CONNECT_STEP_CHECK_RUNNING:
-		if (!do_spawn (&priv->pid, NULL, NULL, error, priv->ipsec_path, "auto", "--status", NULL))
+		if (!do_spawn (self, &priv->pid, NULL, NULL, error, priv->ipsec_path, "auto", "--status", NULL))
 			return FALSE;
 		priv->watch_id = g_child_watch_add (priv->pid, check_running_cb, self);
 		return TRUE;
@@ -1404,7 +1451,7 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 				return FALSE;
 
 			/* Ensure the right IPsec kernel stack is loaded */
-			success = do_spawn (&priv->pid, NULL, NULL, error, stackman_path, "start", NULL);
+			success = do_spawn (self, &priv->pid, NULL, NULL, error, stackman_path, "start", NULL);
 			if (success)
 				priv->watch_id = g_child_watch_add (priv->pid, child_watch_cb, self);
 			return success;
@@ -1415,9 +1462,9 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 	case CONNECT_STEP_IPSEC_START:
 		/* Start the IPsec service */
 		if (priv->openswan)
-			success = do_spawn (&priv->pid, NULL, NULL, error, priv->ipsec_path, "setup", "start", NULL);
+			success = do_spawn (self, &priv->pid, NULL, NULL, error, priv->ipsec_path, "setup", "start", NULL);
 		else {
-			success = do_spawn (&priv->pid, NULL, NULL, error,
+			success = do_spawn (self, &priv->pid, NULL, NULL, error,
 			                    priv->pluto_path, "--config", SYSCONFDIR "/ipsec.conf",
 			                    NULL);
 		}
@@ -1430,14 +1477,14 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 	case CONNECT_STEP_WAIT_READY:
 		if (!priv->retries)
 			priv->retries = 30;
-		if (!do_spawn (&priv->pid, NULL, NULL, error, priv->ipsec_path, "auto", "--ready", NULL))
+		if (!do_spawn (self, &priv->pid, NULL, NULL, error, priv->ipsec_path, "auto", "--ready", NULL))
 			return FALSE;
 		priv->watch_id = g_child_watch_add (priv->pid, child_watch_cb, self);
 		return TRUE;
 
 	case CONNECT_STEP_CONFIG_ADD:
 		g_assert (uuid);
-		if (!do_spawn (&priv->pid, &fd, NULL, error, priv->ipsec_path,
+		if (!do_spawn (self, &priv->pid, &fd, NULL, error, priv->ipsec_path,
 		               "auto", "--replace", "--config", "-", uuid, NULL))
 			return FALSE;
 		priv->watch_id = g_child_watch_add (priv->pid, child_watch_cb, self);
@@ -1449,7 +1496,7 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 
 	case CONNECT_STEP_CONNECT:
 		g_assert (uuid);
-		if (!spawn_pty (&up_stdout, &up_stderr, &up_pty, &priv->pid, error,
+		if (!spawn_pty (self, &up_stdout, &up_stderr, &up_pty, &priv->pid, error,
 		                priv->ipsec_path, "auto", "--up", uuid, NULL))
 			return FALSE;
 		priv->watch_id = g_child_watch_add (priv->pid, child_watch_cb, self);
@@ -1671,13 +1718,13 @@ real_disconnect (NMVpnServicePlugin *plugin, GError **error)
 
 	if (!priv->managed) {
                 const char *uuid = nm_connection_get_uuid (priv->connection);
-		ret = do_spawn (&priv->pid, NULL, NULL, error,
+		ret = do_spawn (plugin, &priv->pid, NULL, NULL, error,
 		                priv->ipsec_path, "auto", "--delete", uuid, NULL);
 	} else if (priv->openswan) {
-		ret = do_spawn (&priv->pid, NULL, NULL, error,
+		ret = do_spawn (plugin, &priv->pid, NULL, NULL, error,
                                 priv->ipsec_path, "setup", "stop", NULL);
 	} else {
-		ret = do_spawn (&priv->pid, NULL, NULL, error,
+		ret = do_spawn (plugin, &priv->pid, NULL, NULL, error,
 		                priv->whack_path, "--shutdown", NULL);
 	}
 
@@ -1768,7 +1815,8 @@ setup_signals (void)
 static void
 quit_mainloop (NMLibreswanPlugin *plugin, gpointer user_data)
 {
-	g_main_loop_quit ((GMainLoop *) user_data);
+	g_signal_handlers_disconnect_by_func (plugin, quit_mainloop, user_data);
+	unblock_quit (plugin);
 }
 
 int
@@ -1821,6 +1869,7 @@ main (int argc, char *argv[])
 
 	plugin = g_initable_new (NM_TYPE_LIBRESWAN_PLUGIN, NULL, &error,
 	                         NM_VPN_SERVICE_PLUGIN_DBUS_SERVICE_NAME, bus_name,
+	                         NM_VPN_SERVICE_PLUGIN_DBUS_WATCH_PEER, !debug,
 	                         NULL);
 	if (!plugin) {
 		g_warning ("Failed to initialize a plugin instance: %s", error->message);
@@ -1849,8 +1898,9 @@ main (int argc, char *argv[])
 
 	loop = g_main_loop_new (NULL, FALSE);
 
+	block_quit (plugin);
 	if (!persist)
-		g_signal_connect (plugin, "quit", G_CALLBACK (quit_mainloop), loop);
+		g_signal_connect (plugin, "quit", G_CALLBACK (quit_mainloop), NULL);
 
 	setup_signals ();
 	g_main_loop_run (loop);
